from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function, Contract
from slither.core.cfg.node import Node
import re


class CurveReadonlyReentrancy(AbstractDetector):
    """
    Sees if a contract has a beforeTokenTransfer function.
    """

    ARGUMENT = "pess-curve-readonly-reentrancy"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "Curve readonly-reentrancy"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/pessimistic-io/slitherin/blob/master/docs/curve_readonly_reentrancy.md"
    WIKI_TITLE = "Curve Readonly Reentrancy"
    WIKI_DESCRIPTION = "Check docs"
    WIKI_EXPLOIT_SCENARIO = "-"
    WIKI_RECOMMENDATION = "Check docs"

    VULNERABLE_FUNCTION_CALLS = ["get_virtual_price", "lp_price"]
    visited = []
    contains_reentrancy_check = {}

    def is_curve_integration(self, c: Contract) -> bool:
        """
        Iterates over all external function calls, and checks the interface/contract name
        for a specific keywords to decide if the contract integrates with balancer

        @note as for now, we are skipping any name/interface checks
        """

        return True

    def _has_reentrancy_check(self, node: Node) -> bool:
        if node in self.visited:
            return self.contains_reentrancy_check[node]

        self.visited.append(node)
        self.contains_reentrancy_check[node] = False

        for _, n in node.high_level_calls:
            if isinstance(n, Function):
                if not n.name:
                    continue
                if n.name == "withdraw_admin_fee":
                    self.contains_reentrancy_check[node] = True
                    return True

        # TODO: currently using regex to find address().call(*.withdraw_admin_fees)
        # It would be better, if it is rewritten in a ast analysis form

        if re.search(r"withdraw_admin_fee", str(node)):
            return True

        has_check = False
        for internal_call in node.internal_calls:
            if isinstance(internal_call, Function):
                for f_node in internal_call.nodes:
                    has_check |= self._has_reentrancy_check(f_node)

        self.contains_reentrancy_check[node] = has_check
        return has_check

    def _check_function(self, function: Function) -> list:
        has_dangerous_call = False
        dangerous_call = None
        for n in function.nodes:
            for c, fc in n.high_level_calls:
                if isinstance(fc, Function):
                    if fc.name in self.VULNERABLE_FUNCTION_CALLS:
                        dangerous_call = n  # Saving only first dangerous call
                        has_dangerous_call = True
                        break

        if has_dangerous_call and not any(
            [self._has_reentrancy_check(node) for node in function.nodes]
        ):
            return [dangerous_call]
        return []

    def _detect(self) -> List[Output]:
        """Main function"""
        result = []
        for contract in self.compilation_unit.contracts_derived:
            if not self.is_curve_integration(contract):
                continue
            res = []
            for f in contract.functions_and_modifiers_declared:
                function_result = self._check_function(f)
                if function_result:
                    res.extend(function_result)
            if res:
                info = [
                    "Curve readonly-reentrancy vulnerability detected in ",
                    contract,
                    ":\n",
                ]
                for r in res:
                    info += [
                        "\tThe answer of ",
                        r,
                        " call could be manipulated through readonly-reentrancy\n",
                    ]
                res = self.generate_result(info)
                res.add(r)
                result.append(res)

        return result



import napalm.package
import slitherin as slitherin


def entry_point():
    """This is the entry point for the napalm package.

    It provisions your detection modules and provides them to napalm.

    It returns a dictionary of Collections, keyed by the name of the collection.
    """
    _include = ("detectors",)

    return [
        collection
        for collection in napalm.package.entry_point(slitherin)
        if collection.collection_name in _include
    ]



from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function
from slither.slithir.operations.event_call import EventCall

VULNERABLE_VERSIONS = ['0.2.15', '0.2.16', '0.3.0']
class CurveVyperReentrancy(AbstractDetector):
    ARGUMENT = 'pess-curve-vyper-reentrancy' # slither will launch the detector with slither.py --detect mydetector
    HELP = f'Vyper compiler versions {", ".join(VULNERABLE_VERSIONS)} are vulnerable to malfunctioning re-entrancy guards. Upgrade your compiler version.'
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = 'https://github.com/pessimistic-io/slitherin/blob/master/docs/curve_vyper_reentrancy.md'
    WIKI_TITLE = 'Vulnerable Vyper version'
    WIKI_DESCRIPTION = "Some Vyper versions are vulnerable to malfunctioning re-entrancy guards."
    WIKI_EXPLOIT_SCENARIO = 'https://hackmd.io/@LlamaRisk/BJzSKHNjn'
    WIKI_RECOMMENDATION = 'Upgrade the version of your Vyper compiler.'

    def _detect(self) -> List[Output]:
        res = []
        if not self.compilation_unit.is_vyper:
            return res
        compiler_version = self.compilation_unit.compiler_version.version
        if compiler_version not in VULNERABLE_VERSIONS:
            return res
        
        for contract in self.contracts:
            for f in contract.functions_entry_points:
                for modifier in f.modifiers:
                    if modifier.name.startswith("nonreentrant"):
                        res += ["\t", modifier.name, " in ", f, " function.\n"]
        if not res:
            return res
        return [self.generate_result(
            [f"Vyper {compiler_version} compiler version is vulnerable to malfunctioning re-entrancy guards. Found vulnerable usages:\n",
             *[x for x in res],
             "\n"
            ]
        )]



import os
import re
from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function
from slither.analyses.data_dependency.data_dependency import is_dependent


from ...consts import ARBITRUM_KEY


class ArbitrumChainlinkPriceFeed(AbstractDetector):
    """
    Checks if sequencer uptime is verified when chainlink price feed is used
    """

    ARGUMENT = "pess-arb-chainlink-price-feed"  # slither will launch the detector with slither.py --detect mydetector

    HELP = "Sequencer uptime is not checked"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/pessimistic-io/slitherin/blob/master/docs/arb_chainlink_price_feed.md"
    WIKI_TITLE = "Arbitrum chainlink sequencer uptime"
    WIKI_DESCRIPTION = "Arbitrum chainlink sequencer uptime check"
    WIKI_EXPLOIT_SCENARIO = "N/A"
    WIKI_RECOMMENDATION = "Check sequencer uptime (see docs)"

    def _contains_latest_round_call(self, f: Function) -> bool:
        ret = set()
        for node in f.nodes:
            for _, v in node.high_level_calls:
                if isinstance(v, Function) and v.name == "latestRoundData":
                    return True
        return False

    def _contains_sequencer_check(self, f: Function) -> bool:
        # Checks if the keyword sequencer used in function.
        # This check might contain FPs. However, since most of devs will copy-paste or get inspiration
        # from the chainlink docs, as for now, I think it should work well.

        for node in f.nodes:
            if re.search(r"sequencer", str(node), re.IGNORECASE):
                return True

        return False

    def _detect(self) -> List[Output]:
        """Main function"""
        results = []

        if os.getenv(ARBITRUM_KEY) is None:
            return results

        for contract in self.compilation_unit.contracts_derived:
            for f in contract.functions_and_modifiers:
                if self._contains_latest_round_call(
                    f
                ) and not self._contains_sequencer_check(f):
                    info = [
                        "Sequencer uptime status is not checked when using price feed in:\n\t",
                        f,
                        "\nCheck https://docs.chain.link/data-feeds/l2-sequencer-feeds for more details\n",
                    ]
                    results.append(self.generate_result(info))
        return results



from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.declarations import Function, Contract
from slither.core.cfg.node import Node


class BalancerReadonlyReentrancy(AbstractDetector):
    """
    Sees if a contract has a beforeTokenTransfer function.
    """

    ARGUMENT = "pess-balancer-readonly-reentrancy"  # slither will launch the detector with slither.py --detect mydetector
    HELP = "Balancer readonly-reentrancy"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/pessimistic-io/slitherin/blob/master/docs/balancer/readonly_reentrancy.md"
    WIKI_TITLE = "Balancer Readonly Reentrancy"
    WIKI_DESCRIPTION = "Check docs"
    WIKI_EXPLOIT_SCENARIO = "-"
    WIKI_RECOMMENDATION = "Check docs"

    VULNERABLE_FUNCTION_CALLS = ["getRate", "getPoolTokens"]
    visited = []
    contains_reentrancy_check = {}

    def is_balancer_integration(self, c: Contract) -> bool:
        """
        Iterates over all external function calls, and checks the interface/contract name
        for a specific keywords to decide if the contract integrates with balancer
        """
        for (
            fcontract,
            _,
        ) in c.all_high_level_calls:
            contract_name = fcontract.name.lower()
            if any(map(lambda x: x in contract_name, ["balancer", "ivault", "pool"])):
                return True

    def _has_reentrancy_check(self, node: Node) -> bool:
        if node in self.visited:
            return self.contains_reentrancy_check[node]

        self.visited.append(node)
        self.contains_reentrancy_check[node] = False

        for c, n in node.high_level_calls:
            if isinstance(n, Function):
                if not n.name:
                    continue
                if (
                    n.name == "ensureNotInVaultContext"
                    and c.name == "VaultReentrancyLib"
                ) or (
                    n.name == "manageUserBalance"
                ):  # TODO check if errors out
                    self.contains_reentrancy_check[node] = True
                    return True

        has_check = False
        for internal_call in node.internal_calls:
            if isinstance(internal_call, Function):
                has_check |= self._has_reentrancy_check(internal_call)
                # self.contains_reentrancy_check[internal_call] |= has_check

        self.contains_reentrancy_check[node] = has_check
        return has_check

    def _check_function(self, function: Function) -> list:
        has_dangerous_call = False
        dangerous_call = None
        for n in function.nodes:
            for c, fc in n.high_level_calls:
                if isinstance(fc, Function):
                    if fc.name in self.VULNERABLE_FUNCTION_CALLS:
                        dangerous_call = n  # Saving only first dangerous call
                        has_dangerous_call = True
                        break

        if has_dangerous_call and not any(
            [self._has_reentrancy_check(node) for node in function.nodes]
        ):
            return [dangerous_call]
        return []

    def _detect(self) -> List[Output]:
        """Main function"""
        result = []
        for contract in self.compilation_unit.contracts_derived:
            if not self.is_balancer_integration(contract):
                continue
            res = []
            for f in contract.functions_and_modifiers_declared:
                function_result = self._check_function(f)
                if function_result:
                    res.extend(function_result)
            if res:
                info = [
                    "Balancer readonly-reentrancy vulnerability detected in ",
                    contract,
                    ":\n",
                ]
                for r in res:
                    info += [
                        "\tThe answer of ",
                        r,
                        " call could be manipulated through readonly-reentrancy\n",
                    ]
                res = self.generate_result(info)
                res.add(r)
                result.append(res)

        return result



from typing import List
from slither.utils.output import Output
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.slithir.operations.high_level_call import HighLevelCall
from slither.slithir.operations.internal_call import InternalCall
from slither.slithir.operations.solidity_call import SolidityCall
from slither.slithir.operations.binary import Binary
from slither.analyses.data_dependency.data_dependency import is_dependent


class PriceManipulationDetector(AbstractDetector):
    ARGUMENT = 'pess-price-manipulation' # slither will launch the detector with slither.py --detect mydetector
    HELP = 'Contract math can be manipulated through external token transfers.'
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.LOW

    WIKI = 'https://github.com/pessimistic-io/slitherin/blob/master/docs/price_manipulation.md'
    WIKI_TITLE = '# Price Manipulation through token transfers'
    WIKI_DESCRIPTION = "Calculations could be manipulated through direct transfers to the contract, increasing its balance as they depends on these balances."
    WIKI_EXPLOIT_SCENARIO = 'N/A'
    WIKI_RECOMMENDATION = 'Avoid possible manipulations of calculations because of external transfers.'

    def _detect(self) -> List[Output]:
        all_balance_vars = []
        all_supply_vars = []
        all_binary_ops = []
        for contract in self.contracts:
            if not contract.is_interface:
                for func in contract.functions:
                    for n in func.nodes:
                        for x in n.irs:
                            if isinstance(x, SolidityCall):
                                if x.function.name == "balance(address)" or x.function.name == "self.balance" or x.function.name == "this.balance()":
                                    all_balance_vars.append((n, x._lvalue))
                            if isinstance(x, HighLevelCall):
                                if str(x.function_name).lower() == "balanceof":
                                    all_balance_vars.append((n, x._lvalue))
                                if "supply" in str(x.function_name).lower():
                                    all_supply_vars.append((n, x._lvalue))
                            if isinstance(x, InternalCall):
                                if "supply" in str(x.function_name).lower():
                                    all_supply_vars.append((n, x._lvalue))
                            if isinstance(x, Binary):
                                all_binary_ops.append((n, x))
        results = []
        for (balance_node, bal) in all_balance_vars:
            for (supply_node, supply) in all_supply_vars:
                for (node, bin_op) in all_binary_ops:
                    l, r = bin_op.get_variable
                    is_bal_dependent = is_dependent(l, bal, contract)
                    is_supply_dependent = is_dependent(r, supply, contract)
                    if is_bal_dependent and is_supply_dependent:
                        results.append((node, balance_node, supply_node))
        if not results:
            return []
        
        
        response = []
        for issue_node, balance_node, supply_node in results:
            res = []
            res.append("Calculation ")
            res.append(issue_node)
            res.append(" depends on balance and token supply - these values could be manipulated through external calls.\n")
            res.append("\tBalance dependency: ")
            res.append(balance_node)
            res.append("\n")
            res.append("\tSupply dependency: ")
            res.append(supply_node)
            res.append("\n")
            response.append(self.generate_result(res))
        return response
